/*
 * Copyright (c) 2022 Libre Solar Technologies GmbH
 *
 * SPDX-License-Identifier: Apache-2.0
 */

#include <zephyr/kernel.h>
#include <zephyr/device.h>
#include <zephyr/drivers/uart.h>
#include <zephyr/logging/log.h>
#include <zephyr/drivers/gpio.h>
#include <zephyr/pm/device.h>

#include <string.h>

LOG_MODULE_REGISTER(app);

#define MSG_SIZE 255

typedef struct uart_irq_rx_buffer_t
{
	char uart0_rx_buf[MSG_SIZE];
	int length;
} uart_irq_rx_buffer_t;

/* queue to store up to 10 messages (aligned to 4-byte boundary) */
K_MSGQ_DEFINE(uart0_msgq, sizeof(uart_irq_rx_buffer_t), 10, 4);
K_MSGQ_DEFINE(uart1_msgq, MSG_SIZE, 10, 4);

// static const struct device *const uart_dev = DEVICE_DT_GET(UART_DEVICE_NODE);
static const struct device *const uart_dev = DEVICE_DT_GET(DT_NODELABEL(uart0));
static const struct device *const uart1_dev = DEVICE_DT_GET(DT_NODELABEL(uart1));

/* Regulator Handles Setup */
#define SENSOR_EN_NODE	DT_ALIAS(sensorpwren)
#if !DT_NODE_HAS_STATUS(SENSOR_EN_NODE, okay)
#error "Unsupported board: gpspwren devicetree alias is not defined"
#endif
static const struct gpio_dt_spec sensor_pwr_en = GPIO_DT_SPEC_GET_OR(SENSOR_EN_NODE, gpios,
							      {0});

#define MPCI_PWR_EN_NODE	DT_ALIAS(mpcipwren)
#if !DT_NODE_HAS_STATUS(MPCI_PWR_EN_NODE, okay)
#error "Unsupported board: gpspwren devicetree alias is not defined"
#endif
static const struct gpio_dt_spec mpci_pwr_en = GPIO_DT_SPEC_GET_OR(MPCI_PWR_EN_NODE, gpios,
							      {0});



/* RS485 Handles Setup */
#define RS485_RE_EN_NODE	DT_ALIAS(rs485reen)
#if !DT_NODE_HAS_STATUS(RS485_RE_EN_NODE, okay)
#error "Unsupported board: gpspwren devicetree alias is not defined"
#endif
static const struct gpio_dt_spec rs485_re_en = GPIO_DT_SPEC_GET_OR(RS485_RE_EN_NODE, gpios,
							      {0});

#define RS485_DE_EN_NODE	DT_ALIAS(rs485deen)
#if !DT_NODE_HAS_STATUS(RS485_DE_EN_NODE, okay)
#error "Unsupported board: gpspwren devicetree alias is not defined"
#endif
static const struct gpio_dt_spec rs485_de_en = GPIO_DT_SPEC_GET_OR(RS485_DE_EN_NODE, gpios,
							      {0});

/* receive buffer used in UART ISR callback */
uart_irq_rx_buffer_t uart0_irq_rx_buffer =
{
	.uart0_rx_buf = {0},
	.length = 0
};
static int uart0_rx_buf_pos = 0;

static char uart1_rx_buf[MSG_SIZE] = {0};
static int uart1_rx_buf_pos = 0;


void my_timer_handler(struct k_timer *dummy)
{
		uart0_irq_rx_buffer.length = uart0_rx_buf_pos;
		uart0_irq_rx_buffer.uart0_rx_buf[uart0_rx_buf_pos] = '\0';

		/* if queue is full, message is silently dropped */
		k_msgq_put(&uart0_msgq, &uart0_irq_rx_buffer, K_NO_WAIT);
		uart0_rx_buf_pos = 0;
}

K_TIMER_DEFINE(my_timer, my_timer_handler, NULL);


/*
 * Read characters from UART until line end is detected. Afterwards push the
 * data to the message queue.
 */
void serial_cb(const struct device *dev, void *user_data)
{
	uint8_t c;

	if (!uart_irq_update(uart_dev)) {
		return;
	}

	while (uart_irq_rx_ready(uart_dev)) {
		if(uart_fifo_read(dev, &c, 1))
		{
			if (uart0_rx_buf_pos < (MSG_SIZE - 1)) {
				uart0_irq_rx_buffer.uart0_rx_buf[uart0_rx_buf_pos++] = c;
			}
			else
			{
				// LOG_ERR("uart0_irq_rx_buffer FULL, chars dropped!");
			}
		}
	}

	// This timer acts as an inter char timeout delay.
	/* we stop the previous running one shot timer */
	k_timer_stop(&my_timer);
	/* start a one-shot timer that expires after X us */
	k_timer_start(&my_timer, K_USEC(31), K_NO_WAIT);
}

//

/*
 * Read characters from UART until line end is detected. Afterwards push the
 * data to the message queue.
 */
void uart1_serial_cb(const struct device *dev, void *user_data)
{
	uint8_t c;

	if (!uart_irq_update(dev)) {
		return;
	}

	while (uart_irq_rx_ready(dev)) {

		if (uart_fifo_read(dev, &c, 1) == 0)
		{
			uart1_rx_buf[uart1_rx_buf_pos] = '\0';

			/* if queue is full, message is silently dropped */
			k_msgq_put(&uart1_msgq, &uart1_rx_buf, K_NO_WAIT);
			uart1_rx_buf_pos = 0;

		}
		else if (uart1_rx_buf_pos < (sizeof(uart1_rx_buf) - 1)) {
			uart1_rx_buf[uart1_rx_buf_pos++] = c;
		}
	}
}

/*
 * Print a null-terminated string character by character to the UART interface
 */
void print_uart(const struct device *dev, char *buf)
{
	int msg_len = strlen(buf);

	for (int i = 0; i < msg_len; i++) {
		uart_poll_out(dev, buf[i]);
	}
}

int sensor_pwr_en_set(bool val)
{
    int err;

    err = gpio_pin_set_dt(&sensor_pwr_en, val);
	if (err != 0) {
		LOG_ERR("Error %d: failed to write %s pin %d\n",
		       err, sensor_pwr_en.port->name, sensor_pwr_en.pin);
		return err;
	}

	return err;
}

int mpci_pwr_en_set(bool val)
{
    int err;

    err = gpio_pin_set_dt(&mpci_pwr_en, val);
	if (err != 0) {
		LOG_ERR("Error %d: failed to write %s pin %d\n",
		       err, mpci_pwr_en.port->name, mpci_pwr_en.pin);
		return err;
	}

	return err;
}



int rs485_re_en_set(bool val)
{
    int err;

    err = gpio_pin_set_dt(&rs485_re_en, !val);
	if (err != 0) {
		LOG_ERR("Error %d: failed to write %s pin %d\n",
		       err, rs485_re_en.port->name, rs485_re_en.pin);
		return err;
	}

	return err;
}

int rs485_de_en_set(bool val)
{
    int err;

    err = gpio_pin_set_dt(&rs485_de_en, !val);
	if (err != 0) {
		LOG_ERR("Error %d: failed to write %s pin %d\n",
		       err, rs485_de_en.port->name, rs485_de_en.pin);
		return err;
	}

	return err;
}


void main(void)
{
	int err = 0;

	printk("Entering main\r\n");

	// Sensor PWR Enable
    if (!device_is_ready(sensor_pwr_en.port)) {
		LOG_ERR("Error: PWM device %s is not ready\n",
		       sensor_pwr_en.port->name);
	}

    err = gpio_pin_configure_dt(&sensor_pwr_en, GPIO_OUTPUT);
	if (err != 0) {
		LOG_ERR("Error %d: failed to configure %s pin %d\n",
		       err, sensor_pwr_en.port->name, sensor_pwr_en.pin);
	}

	// MPCI PWR Enable
	if (!device_is_ready(mpci_pwr_en.port)) {
		LOG_ERR("Error: PWM device %s is not ready\n",
		       mpci_pwr_en.port->name);
	}

    err = gpio_pin_configure_dt(&mpci_pwr_en, GPIO_OUTPUT);
	if (err != 0) {
		LOG_ERR("Error %d: failed to configure %s pin %d\n",
		       err, mpci_pwr_en.port->name, mpci_pwr_en.pin);
	}

	// RS485 DE PIN
	if (!device_is_ready(rs485_de_en.port)) {
		LOG_ERR("Error: PWM device %s is not ready\n",
		       rs485_de_en.port->name);
	}

    err = gpio_pin_configure_dt(&rs485_de_en, GPIO_OUTPUT);
	if (err != 0) {
		LOG_ERR("Error %d: failed to configure %s pin %d\n",
		       err, rs485_de_en.port->name, rs485_de_en.pin);
	}

	// RS485 RE PIN
    if (!device_is_ready(rs485_re_en.port)) {
		LOG_ERR("Error: PWM device %s is not ready\n",
		       rs485_re_en.port->name);
	}

    err = gpio_pin_configure_dt(&rs485_re_en, GPIO_OUTPUT);
	if (err != 0) {
		LOG_ERR("Error %d: failed to configure %s pin %d\n",
		       err, rs485_re_en.port->name, rs485_re_en.pin);
	}

	// Puts RS485 chip into sleep mode
	rs485_de_en_set(false);
	rs485_re_en_set(true);

	// Powers on the RS232 driver for transmission
	mpci_pwr_en_set(true);
	sensor_pwr_en_set(false);

	// Wait for the power to stabilise
	k_sleep(K_SECONDS(1));



	if (!device_is_ready(uart_dev)) {
		printk("uart_dev not found!");
	}

	/* configure interrupt and callback to receive data */
	uart_irq_callback_user_data_set(uart_dev, serial_cb, NULL);
	uart_irq_rx_enable(uart_dev);



	if (!device_is_ready(uart1_dev)) {
		printk("uart1_dev not found!");
	}

	/* configure interrupt and callback to receive data */
	uart_irq_callback_user_data_set(uart1_dev, uart1_serial_cb, NULL);
	uart_irq_rx_enable(uart1_dev);

	k_sleep(K_SECONDS(1));

	// Transmit a message
	print_uart(uart_dev, "AT\r\n");
	printk("AT\r\n");

	// If I attach something like a modem simulator (to UART0) and get a response
	// then the power goes down to 98uA, but if I just connect to a terminal
	// that does not respond back then the power is almost 1mA. So it seems
	// I have to receive data on UART0 before powering down, then the power down
	// works as exected.

	k_sleep(K_SECONDS(1));


	/* Power down the UARTs */
    err = pm_device_action_run(uart_dev, PM_DEVICE_ACTION_SUSPEND);
    if (err) {
        printk("pm_device_action_run() failed");
    }

	err = pm_device_action_run(uart1_dev, PM_DEVICE_ACTION_SUSPEND);
    if (err) {
        printk("pm_device_action_run() failed");
    }

	// Power down the RS232 driver
	mpci_pwr_en_set(false);

	// Just spin forever
	for(;;)
	{
		k_sleep(K_SECONDS(1));
	}
}
